import Player from "./player.js"
import Input from "./input.js"
import Event from "./event.js"
import Message from "./message.js"

import { average } from "./util.js"

export default class Game {
  constructor(canvas) {
    this.canvas = canvas
    this.ctx = canvas.getContext("2d")
    this.timestamp = 0
    this.player = new Player(this, 500, 450)
    this.actors = [this.player]

    this.input = new Input().initialize()

    this.currentRoom = null
    this.events = {
      "log_test": new Event("log_test", () => console.log("Log events work!")),
      "change_color": new Event("change_color", object => object.setProperty("color", "blue")),
      "show_message": new Event("show_message", this.openMessage.bind(this)),
      "transport": new Event("transport", this.initiateTransport.bind(this))
    }

    this.message = null

    this.fpsBuffer = [60, 60, 60, 60, 60, 60, 60, 60, 60, 60]
  }

  triggerEvent(eventName, object) {
    const event = this.events[eventName]
    if (event) event.trigger(object)
    else console.error("Unknown event " + eventName)
  }

  start() {
    this.loadRoom(this.assets.get("sampleRoom"))
    requestAnimationFrame(this.loop.bind(this))
  }

  loadRoom(room) {
    this.currentRoom = room
    this.currentRoom.objects.forEach(roomObject => this.actors.push(roomObject))
  }

  openMessage(object) {
    this.message = new Message(this, object.getProperty("messageText"))
  }

  closeMessage(message) {
    this.message = null
  }

  initiateTransport(object) {
    const coords = JSON.parse(object.getProperty("transport_to_coords"))
    this.actors = [this.player]
    this.loadRoom(this.assets.get(object.getProperty("transport_to_room")))
    this.player.x = coords[0]
    this.player.y = coords[1]
  }

  loop(timestamp) {
    this.dt = timestamp - this.timestamp
    const fps = 1000 / this.dt
    this.fpsBuffer.pop()
    this.fpsBuffer.unshift(fps)
    this.timestamp= timestamp
    this.tick(this.dt)
    this.draw()

    requestAnimationFrame(this.loop.bind(this))
  }

  tick(dt) {
    if (this.message) {
      this.message?.tick(dt)
    } else {
      this.actors.forEach(actor => actor.tick(dt))
      Object.values(this.events).forEach(e => e.nextFrame())
    }
    this.input.tick()
  }

  draw() {
    const { canvas, ctx } = this
    ctx.fillStyle = "black"
    ctx.fillRect(0, 0, canvas.width, canvas.height)
    this.currentRoom.draw(ctx)
    this.actors.forEach(actor => actor.draw(ctx))
    this.message?.draw(ctx)
    this.drawFps(ctx)
    this.drawPlayerPosition(ctx)
  }

  drawFps(ctx) {
    ctx.fillStyle = "white"
    ctx.fillRect(ctx.canvas.width, 0, -25, 20)
    ctx.strokeStyle = "black"
    ctx.textBaseline = "top"
    ctx.textAlign = "right"
    ctx.font = "bold 20px serif"
    ctx.fillText(Math.round(average(this.fpsBuffer)), ctx.canvas.width, 0)
    ctx.strokeText(Math.round(average(this.fpsBuffer)), ctx.canvas.width, 0)
  }

  drawPlayerPosition(ctx) {
    if (!this.player) return

    ctx.fillStyle = "white"
    ctx.fillRect(ctx.canvas.width, 20, -80, 20)
    ctx.strokeStyle = "black"
    ctx.textBaseline = "top"
    ctx.textAlign = "right"
    ctx.font = "bold 20px serif"
    ctx.fillText(`${this.player.x}, ${this.player.y}`, ctx.canvas.width, 20)
    ctx.strokeText(`${this.player.x}, ${this.player.y}`, ctx.canvas.width, 20)
  }
}
