import RoomObject from "./roomObject.js"
import { doRectanglesOverlap } from "./util.js"

export default class Room {
  constructor(game, json, name) {
    this.game = game
    this.json = json
    this.name = name
    const objectJson = this.objectLayers.map(layer => layer.objects).flat()
    this.objects = objectJson.map(RoomObject.fromJson.bind(null, this.game))
  }

  get tilesetsToLoad() {
    const ts = {}
    this.json.tilesets.forEach((tileset, index) => {
      ts[`${this.name}-${index}`] = tileset.source
    })
    return ts
  }

  populateTilesets(assets) {
    this.tilesets = this.json.tilesets.map((tileset, index) => {
      const ts = assets.get(`${this.name}-${index}`)
      ts.populateImage(assets)
      return ts
    })
  }

  draw(ctx) {
    this.tileLayers.forEach(this.drawTileLayer.bind(this, ctx))
  }

  drawLayer(ctx, layer) {
    if (layer.type == "tilelayer") this.drawTileLayer(ctx, layer)
  }

  get tileLayers() {
    return this.json.layers.filter(layer => layer.type == "tilelayer")
  }

  get objectLayers() {
    return this.json.layers.filter(layer => layer.type == "objectgroup")
  }

  objectsUnderRectangle(rect) {
    return this.objects.filter(object => doRectanglesOverlap(object, rect))
  }

  tilesUnderRectangle(rect) {
    return this.tileLayers.map(layer => this.tilesUnderRectangleInLayer(layer, rect)).flat()
  }
  
  tilesUnderRectangleInLayer(layer, rect) {
    return [{ x: rect.x, y: rect.y },
	    { x: rect.x + rect.width, y: rect.y },
	    { x: rect.x, y: rect.y + rect.height },
	    { x: rect.x + rect.width, y: rect.y + rect.height }
	   ].map(point => {
	     const tileset = this.tilesets[0]
	     const { x, y } = point
	     const tileX = Math.floor(x / tileset.tileWidth)
	     const tileY = Math.floor(y / tileset.tileHeight)
	     const index = tileX + (tileY * layer.width)
	     const tileIndex = layer.data[index] - 1
	     return tileset.tileAt(tileIndex)
	   })
  }

  drawTileLayer(ctx, layer) {
    for (let y = 0; y < layer.height; y++) {
      for (let x = 0; x < layer.width; x++) {
	const index = x + (y * layer.width)
	let tileIndex = layer.data[index] + 1
	if (tileIndex == 1) continue
	let tileset = this.json.tilesets
	    .sort((a, b) => b.firstgid - a.firstgid)
	    .find(ts => ts.firstgid <= tileIndex)
	const { firstgid } = tileset
	tileset = this.tilesets.find(ts => ts.source == tileset.source)
	tileIndex -= firstgid + 1

	const [sx, sy] = tileset.tileOffset(tileIndex)
	ctx.drawImage(
	  tileset.image,
	  sx,
	  sy,
	  tileset.tileWidth,
	  tileset.tileHeight,
	  x * this.json.tilewidth,
	  y * this.json.tileheight,
	  this.json.tilewidth,
	  this.json.tileheight
	)
	if (tileset.collides(tileIndex)) {
	  ctx.fillStyle = "#aa660088"
	  ctx.fillRect(
	  x * this.json.tilewidth,
	  y * this.json.tileheight,
	  this.json.tilewidth,
	  this.json.tileheight
	  )
	}
      }
    }
  }
}
